use generic_array::GenericArray;

use crate::hash::{ffi::SHA2Result, sha2::sealed::SHA2Mode};

pub const SHA2_256_HASH_BITS: usize = 256;
pub const SHA2_256_HASH_BYTES: usize = SHA2_256_HASH_BITS / 8;
pub const SHA2_256_HASH_PARTS: usize = SHA2_256_HASH_BYTES / 4 /* storing with u32 */;
pub const SHA2_256_BLOCK_SIZE: usize = 2 * SHA2_256_HASH_BYTES;

pub type Digest256 = [u32; SHA2_256_HASH_PARTS];
pub(crate) const DIGEST256_EMPTY: Digest256 = [0; SHA2_256_HASH_PARTS];

pub struct Mode256;
mod sealed {
    use generic_array::{ArrayLength, GenericArray};

    use super::Mode256;
    pub trait SHA2Mode {
        type HashBits: ArrayLength;
        type HashBytes: ArrayLength;
        type HashParts: ArrayLength;
        type BlockSize: ArrayLength;
        type Digest: AsRef<[u32]> + AsMut<[u32]>;
        type Block: AsRef<[u8]> + AsMut<[u8]>;
    }
    impl SHA2Mode for Mode256 {
        type HashBits = generic_array::typenum::U256;
        type HashBytes = generic_array::typenum::U32;
        type HashParts = generic_array::typenum::U8;
        type BlockSize = generic_array::typenum::U512;
        type Digest = GenericArray<u32, Self::HashParts>;
        type Block = GenericArray<u8, Self::BlockSize>;
    }
    impl Mode256 {}
}

#[derive(Debug, PartialEq, Eq, Clone, Hash)]
pub struct SHA2Context<M: sealed::SHA2Mode> {
    intermediate_hash: GenericArray<u32, M::HashParts>,
    block: GenericArray<u8, M::BlockSize>,
    block_idx: usize,
}

impl<M: sealed::SHA2Mode> SHA2Context<M> {
    pub fn new() -> Self {
        Self {
            intermediate_hash: Default::default(),
            block: Default::default(),
            block_idx: Default::default(),
        }
    }

    pub fn clear(&mut self) {
        *self = Self::new();
    }

    pub fn input(&mut self, data: &[u8]) -> Result<(), SHA2Result> {
        todo!()
    }

    pub fn finish(self) -> Result<M::Digest, SHA2Result> {
        todo!()
    }
}

impl<M: sealed::SHA2Mode> Default for SHA2Context<M> {
    fn default() -> Self {
        Self::new()
    }
}

pub fn sha2_oneshot<M: sealed::SHA2Mode>(data: &[u8]) -> Result<M::Digest, SHA2Result> {
    todo!()
}

pub fn sha2_256_oneshot(data: &[u8]) -> Result<Digest256, SHA2Result> {
    sha2_oneshot::<Mode256>(data).map(|dig| dig.into_array())
}

#[cfg(test)]
mod test {
    use super::*;

    const TEST_VALUES: &[(&str, Digest256)] = &[
        (
            "AAAA",
            [
                0x63c1dd95, 0x1ffedf6f, 0x7fd968ad, 0x4efa39b8, 0xed584f16, 0x2f46e715, 0x114ee184,
                0xf8de9201,
            ],
        ),
        (
            "BAAA",
            [
                0x49e3cd45, 0x27c96cdc, 0x010160ff, 0x08520e0c, 0xb63c6ef8, 0xc4e7d486, 0x08995343,
                0x7f83a159,
            ],
        ),
    ];

    #[test]
    fn test_check() {
        for (input, expected_output) in TEST_VALUES.iter().copied() {
            assert_eq!(sha2_256_oneshot(input.as_bytes()), Ok(expected_output))
        }
    }
}
